


-- Globals in this file:


-- global function Bezier_curve(integer points, sequence p1, sequence p2,
--     return curve

-- global function Bresenham_line(sequence p1, sequence p2)
--     return line

-- global function sightline(sequence p1, sequence p2)
--     return line





-- Return a sequence of xy pairs describing a Bezier curve.
-- The length of the sequence is given by the integer 'points'.
-- Note that this is still a quite naive, direct implementation
-- of the math. I'll see if I can find some way to opimize it.
-------------------------------------------------------------------------------
global function Bezier_curve(integer points, sequence p1,  sequence p2,
					     sequence p3,  sequence p4)
    sequence curve
    atom t, tsq, omt, omtsq, delta, a, b, c, d

    curve = repeat({0,0}, points)
    delta = 1 / (points - 1)
    t = 0.0
    for i = 1 to points do
	omt = 1.0 - t
	omtsq = omt * omt
	tsq = t * t
	a = omtsq * omt
	b = 3 * omtsq * t
	c = 3 * omt * tsq
	d = t * tsq
	curve[i] = {p1[1] * a + p2[1] * b + p3[1] * c + p4[1] * d,
		    p1[2] * a + p2[2] * b + p3[2] * c + p4[2] * d}
	t += delta
    end for
    return curve
end function



-- Return a sequence of xy pairs in an ordinary Bresenham line.
-- The length of the sequence is equal to 1 + the longest of the
-- delta values derived from p1 and p2; i.e. either dx + 1 or dy + 1.
-------------------------------------------------------------------------------
global function Bresenham_line(sequence p1, sequence p2)
    sequence line
    integer  x1, y1, x2, y2, dx, dy, e, xs, ys, count

    x1 = p1[1]
    y1 = p1[2]
    x2 = p2[1]
    y2 = p2[2]
    dx = (x2 - x1)
    dy = (y2 - y1)
    if dx > 0 then xs = 1 else dx = -dx  xs = -1 end if
    if dy > 0 then ys = 1 else dy = -dy  ys = -1 end if
    count = 1
    if dx >= dy then
	line = prepend(repeat({}, dx), {x1, y1})
	e = dx
	dx += dx
	dy += dy
	for x = x1 + xs to x2 by xs do
	    e += dy
	    if e >= dx then
		e -= dx
		y1 += ys
	    end if
	    count += 1
	    line[count] = {x, y1}
	end for
    else
	line = prepend(repeat({}, dy), {x1, y1})
	e = dy
	dx += dx
	dy += dy
	for y = y1 + ys to y2 by ys do
	    e += dx
	    if e >= dy then
		e -= dy
		x1 += xs
	    end if
	    count += 1
	    line[count] = {x1, y}
	end for
    end if
    return line
end function



-- Return a sequence of all the xy pairs that a logical line from p1 to p2
-- would cross through. The length of the sequence is equal to 1 + the
-- Manhatten distance traversed by the line, i.e. dx + dy + 1.
-- Note that if you call it using 2d coordinates: p1 = {x,y}, p2 = {x2,y2},
-- you'll get a 2d sightline out. However, you can also call it with 3d
-- coordinates: p1 = {x,y,z}, p2 = {x2,y2,z2} and get a 3d sightline.
-- The 3d version is often used to find out wheter or not different
-- military units on a topgraphical map can see eachother.
-------------------------------------------------------------------------------
global function sightline(sequence p1, sequence p2)
    sequence line
    integer  x1, y1, z1, x2, y2, z2, dx, dy, dz, dd, e, ez, xs, ys, zs, zf, count

    x1 = p1[1]
    y1 = p1[2]
    x2 = p2[1]
    y2 = p2[2]
    dx = x2 - x1
    dy = y2 - y1
    if dx > 0 then xs = 1 else dx = -dx  xs = -1 end if
    if dy > 0 then ys = 1 else dy = -dy  ys = -1 end if
    count = 1
    if length(p1) = 2 then
	zf = 0
	line = prepend(repeat({}, dx + dy), {x1, y1})
    else
	zf = 1
	z1 = p1[3]
	z2 = p2[3]
	dz = z2 - z1
	if dz > 0 then zs = 1 else dz = -dz  zs = -1 end if
	line = prepend(repeat({}, dx + dy), {x1, y1, z1})
    end if
    if zf then
	dd = dx + dy
	ez = floor(dd/2)
    end if
    if dx >= dy then
	e = dx
	dx += dx
	dy += dy
	for x = x1 + xs to x2 by xs do
	    e += dy
	    if not zf then
		if e >= dx then
		    e -= dx
		    count += 1
		    line[count] = {x, y1}
		    y1 += ys
		end if
		count += 1
		line[count] = {x, y1}
	    else
		if e >= dx then
		    ez += dz
		    while ez >= dd do
			ez -= dd
			z1 += zs
		    end while
		    e -= dx
		    count += 1
		    line[count] = {x, y1, z1}
		    y1 += ys
		end if
		ez += dz
		while ez >= dd do
		    ez -= dd
		    z1 += zs
		end while
		count += 1
		line[count] = {x, y1, z1}
	    end if
	end for
    else
	e = dy
	dx += dx
	dy += dy
	for y = y1 + ys to y2 by ys do
	    e += dx
	    if not zf then
		if e >= dy then
		    e -= dy
		    count += 1
		    line[count] = {x1, y}
		    x1 += xs
		end if
		count += 1
		line[count] = {x1, y}
	    else
		if e >= dy then
		    ez += dz
		    while ez >= dd do
			ez -= dd
			z1 += zs
		    end while
		    count += 1
		    line[count] = {x1, y, z1}
		end if
		ez += dz
		while ez >= dd do
		    ez -= dd
		    z1 += zs
		end while
		count += 1
		line[count] = {x1, y, z1}
	    end if
	end for
    end if
    return line
end function

